import math
import random
import socket
from select import select
import struct
import time
try:
    import renpy
except Exception, e:
    renpy = None

logfile = None # If true, all packets will be saved to it.
verbose = False
server = None

def enable_logging():
    global logfile
    logfile = file("udp.log","ab")
    logfile.write("log@"+struct.pack('<L',int(time.time())))

def log(pkt, addr, now):
    if logfile:
        ent = pkt+repr(addr)
        logfile.write("udp "+struct.pack("<LL",int(now),len(ent))+ent)

console = []
def out(s):
    global console
    console.append(s)
    console = console[-200:]
    if renpy:
        renpy.exports.restart_interaction()

def check_hopping(mob, t, newx, newy, caption):
    age = t-mob.updated
    predicted_x = mob.x+(mob.xvel*age)
    predicted_y = mob.y+(mob.yvel*age)
    delt = abs(predicted_x-newx)+abs(predicted_y-newy)
    if delt>1:
        print "Mob off target by",delt, caption

def packbits(seq):
    a = seq[0] and 1 or 0
    a += seq[1] and 2 or 0
    a += seq[2] and 4 or 0
    a += seq[3] and 8 or 0
    a += seq[4] and 0x10 or 0
    a += seq[5] and 0x20 or 0
    a += seq[6] and 0x40 or 0
    a += seq[7] and 0x80 or 0
    return chr(a)

def unpackbits(c):
    c = ord(c)
    return (c&1, c&2, c&4, c&8, c&0x10, c&0x20, c&0x40, c&0x80)

def sanitized(s):
    ret = ''
    for c in s:
        oc = ord(c)
        try:
            if c in '[{':
                ret += unichr(0x3010)
            elif oc<32 or (oc>=0xd800 and oc<=0xdfff):
                ret += unichr(0xfffd)
            else:
                ret += c
        except Exception,e:
            ret += unichr(0xfffc)
    return ret

class State(object):
    bg_seed = 0L

    def __init__(state):
        state.reset()

    def reset(state):
        state.last = state.epoch = time.time()
        state.nexttick = state.epoch + .01
        state.backgrounds = []
        state.zone = []
        state.dots = []
        for y in range(0,65):
            state.dots.append([False,]*65)

    def long_to_time(self, l):
        # print "long-to-time",l,'epoch=',self.epoch,'res=', self.epoch + (l*TMUL)
        return self.epoch + (l*TMUL)

    def remove_mob(state, uid):
        zone = state.zone
        for i in range(0, len(zone)):
            if zone[i].uid==uid:
                ret = zone[i]
                del zone[i]
                return ret
        return False

    def time_to_long(self, t):
        # print 'time_to_long t =',t,'epoch =',self.epoch, 'res =',(t-self.epoch) * 1024
        return (t-self.epoch) * 1024

class ServerState(State):
    peers = {}
    peersfrom = {}

    name_serial = 0

    def reset(state):
        State.reset(state)
        state.name_serial = 0
        state.peers.clear()
        state.peersfrom.clear()
        # Dots didn't work the way I wanted, so
        #numdots = 1024
        #while numdots>0:
        #    numdots -= state.dispense_dots(random.randint(2, max(random.randint(0,numdots/10),4)))

    def tell(state, mob, text):
        if mob.uid in state.peersfrom:
            sock.sendto('Said\0\0\0\0'+text, state.peersfrom[mob.uid])

    def __init__(state):
        State.__init__(state)
        state.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        state.sock.bind(("0.0.0.0", 18116))
        state.sockfd = state.sock.fileno()

    def send_all(state, addr, now):
        state.sock.sendto("NOW="+struct.pack("<LQ", state.time_to_long(now), state.bg_seed), addr)
        for mob in state.zone:
            state.sock.sendto(mob.state_str(), addr)
            state.sock.sendto(mob.info_str(), addr)
            #if mob.nimi:
            #    state.sock.sendto(mob.name_str(), addr)
        #state.sock.sendto(state.dots_str(), addr)

    def dots_str(state):
        ret = 'DOTS'
        for line in state.dots:
            for i in range(0,64,8):
                ret += packbits(line[i:i+8])
        return ret

    def dispense_dots(self, maxdots):
        starting_maxdots = maxdots
        y = random.randint(0,63)
        y |= 1
        x = random.randint(3,63)
        x |= 1
        while maxdots:
            xv,yv = random.choice(((-1,0),(0,-1),(1,0),(0,1)))
            while maxdots:
                if not self.dots[y][x]:
                    maxdots -= 1
                    self.dots[y][x] = 1
                y += yv
                if (y<1) or (y>61):
                    return starting_maxdots - maxdots
                x += xv
                if (x<1) or (x>61):
                    return starting_maxdots - maxdots
                if not random.randint(0,13):
                    break
        return maxdots

    def broadcast(state, s):
        for addr in state.peersfrom:
            #print "Broadcasting",repr(s),addr
            state.sock.sendto(s, addr)

    def unique_name(state, s):
        for mob in state.zone:
            if mob.nimi==s:
                s = s+str(state.name_serial)
                state.name_serial += 1
        return s

armor_for_shell = {
    0: 0,
    16: 3,
    17: 7,
    18: 5,
    19: 11,
    }

class Mob(object):
    xvel = yvel = 0
    x = y = 0
    updated = 0
    uid = 0

    nimi = None
    last_text = None

    spr = None
    spritenum = 0
    color = (255,255,255)

    size = 32
    stomach_capacity = 0
    stomach_full = 0
    hp = 5
    maxhp= 5

    shell = 0
    shell_size = 32

    next_decision = 0

    def __repr__(self):
        return '<Mob '+repr(self.uid)+'>'

    def init(self, updated):
        self.x = random.randint(0,4096)
        self.y = random.randint(0,4096)
        self.uid = "mo"+struct.pack("<H",random.randint(0,65535))
        self.updated = updated
        self.size = random.randint(16,64)

    def armor(mob):
        return armor_for_shell.get(mob.shell, 5)

    def is_shell(mob):
        return mob.spritenum>=16 and mob.spritenum<128

    def info_str(self):
        ret = 'Info'+self.uid+struct.pack("<BBBBHHHHHHB",
            self.spritenum, self.color[0], self.color[1], self.color[2],
            self.size, self.stomach_capacity, self.stomach_full, self.hp, self.maxhp,
            self.shell_size, self.shell
            )
        if self.nimi:
            ret += self.nimi
        return ret

    def name_str(self):
        return 'Nimi'+self.uid+str(self.nimi)

    def state_str(self):
        return 'MOVU'+self.uid+struct.pack("<LHHhh", state.time_to_long(self.updated), self.x, self.y, int(self.xvel*16), int(self.yvel*16))

    def update_position(mob, now):
        amt = now-mob.updated
        mob.x = mob.x+(mob.xvel*amt)
        mob.y = mob.y+(mob.yvel*amt)

    def decision(mob, now):
        mob.next_decision = now+1024.

    def vec_towards(mob, other):
        return other.x-mob.x, other.y-mob.y

class Shell(Mob):
    spritenum = 0

    def __init__(self, spritenum):
        self.spritenum = spritenum
        self.size = random.randint(32,256)
        self.color = random.randint(44,255), random.randint(44,255), random.randint(44,255)

class Player(Mob):
    size = 32.
    stomach_capacity = 2.
    stomach_full = 0
    hp = 32
    maxhp = 32
    spritenum = 1

    def __init__(self, uid, addr, updated):
        Mob.__init__(self)
        self.addr = addr
        self.uid = uid
        self.updated = updated
        self.x = random.randint(100,3096)
        self.y = random.randint(100,3096)
        self.color = (random.randint(192,255), random.randint(192,255), random.randint(192,255))

class Vihollinen(Mob):
    size = 64.
    stomach_capacity = 20
    next_decision = 0.
    target = None
    color = (255,0,0)
    spritenum = 128

    def init(self, updated):
        Mob.init(self, updated)
        self.uid = 'hi'+self.uid[2:]

    def bite(fish, target, now):
        damage = random.randint(2, int(fish.size*.3))
        damage -= target.armor()
        print "Debug: Biting for",damage
        if damage<=0:
            state.tell(target, "Your armor protects you.")
            fish.wander(now)
            return
        target.hp -= damage
        print "Debug: target hp",target.hp
        fish.stomach_full = min(fish.stomach_capacity, fish.stomach_full+damage)
        if fish.stomach_full>=fish.stomach_capacity:
            fish.wander(now)
        if target.hp<=0:
            state.tell(target,"You are dead.")
            server_kill(target, now)
        state.broadcast(target.info_str())

    def wander(self, now):
        angle_radians = math.radians(random.randint(0,360))
        self.xvel = 16*math.sin(angle_radians)
        self.yvel = 16*math.cos(angle_radians)
        self.next_decision = now+45.

    def decision(self, now):
        if self.target and self.target.hp>0:
            new_mdist = (abs(self.x-self.target.x)+abs(self.y-self.target.y))*.5
            new_target = self.target
        else:
            new_mdist = 40960000
            new_target = None

        for mob in state.zone:
            if isinstance(mob, Player):
                mdist = abs(mob.x-self.x)+abs(mob.y-self.y)
                if mdist<new_mdist and mob.hp>0:
                    new_mdist = mdist
                    new_target = mob
        self.next_decision = now+5.
        if new_target:
            if new_mdist>10:
                target_xdist = new_target.x-self.x
                target_ydist = new_target.y-self.y
                angle_radians = math.atan2(target_xdist, target_ydist)
                self.xvel = 32*math.sin(angle_radians)
                self.yvel = 32*math.cos(angle_radians)
                if new_mdist<200:
                    self.next_decision = now+.1
                else:
                    self.next_decision = max(min(now+(target_ydist/max(self.yvel,1)), now+.1),now+.1)
            else:
                self.xvel = self.yvel = 0
                self.next_decision = now+1.
                self.bite(new_target, now)
            self.target = new_target
            state.broadcast(self.state_str())

def server_kill(target,now):
    target.spritenum = 3
    target.hp = 0
    target.xvel = 0
    target.yvel = 0
    server_drop_shell(target, now)

def update_simulation(state,now):
    state.nexttick = now+.01
    for mob in state.zone:
        age = now-mob.updated
        mob.x += (mob.xvel*age)
        mob.y += (mob.yvel*age)
        if mob.x<0:
            mob.x = 0
            mob.xvel = 0
        elif mob.x>4096:
            mob.x = 4096
            mob.xvel = 0
        if mob.y<0:
            mob.y = 0
            mob.yvel = 0
        elif mob.y>4096:
            mob.y = 4096
            mob.yvel = 0
        #if mob.xvel or mob.yvel:
        #    print 'mob',mob, age, mob.x, mob.y, mob.xvel, mob.yvel
        if now>=mob.next_decision:
            mob.decision(now)
        mob.updated = now
    state.last = now

VELMUL = 1./16.
TMUL = 1./1024.

def long_to_time_OLD(l):
    # Convert a network-format long to a float of seconds since the server
    # epoch.
    return epoch+(l*TMUL)

def time_to_long_OLD(t):
    ret = int((t-state.epoch)*1024)
    print 'ttl',t,ret
    return ret

def init_server():
    global state
    state = ServerState()
    state.bg_seed = random.randint(0,0xffffffff)+(random.randint(0,0xffffffff)<<32)

    now = time.time()

    for i in range(0,20):
        mob = Mob()
        mob.init(now)
        state.zone.append(mob)

    for i in range(0,20):
        mob = Shell(16)
        mob.init(now)
        state.zone.append(mob)
    for i in range(0,10):
        mob = Shell(17)
        mob.init(now)
        state.zone.append(mob)
    for i in range(0,15):
        mob = Shell(18)
        mob.init(now)
        state.zone.append(mob)
    for i in range(0,15):
        mob = Shell(19)
        mob.init(now)
        state.zone.append(mob)
    for i in range(0,6):
        hai = Vihollinen()
        hai.init(now)
        state.zone.append(hai)



def server_loop():
    global server
    server = True
    while 1:
        readable, discard, discard = select([state.sockfd],[],[],.001)
        now = time.time()
        if now>state.nexttick:
            update_simulation(state, now)
        if not readable:
            continue
        data, addr = state.sock.recvfrom(1024)
        log(data,addr,now)
        if data.startswith('CONN'):
            peer = data[4:8]
            rapunimi = state.unique_name(data[8:])
            if peer not in state.peers:
                state.peers[peer] = Player(peer, addr, now)
                state.peers[peer].nimi = rapunimi
                state.zone.append(state.peers[peer])
                if rapunimi:
                    print "DBG: Broadcasting name",repr(rapunimi)
                    state.broadcast('Nimi'+peer+rapunimi)
                state.broadcast(state.peers[peer].state_str())
                state.broadcast(state.peers[peer].info_str())
            else:
                del state.peersfrom[addr]
                state.peers[peer].addr = addr
                if rapunimi!=state.peers[peer].nimi:
                    state.broadcast("Nimi"+peer+rapunimi)
                    state.peers[peer].nimi = rapunimi
            state.peersfrom[addr] = state.peers[peer]
            state.send_all(addr, now)
        elif data.startswith('Drop') and len(data)>=9:
            server_parse_drop(data, addr, now)
        elif data == 'exec':
            server_do_exec()
        elif data.startswith('MOVE'):
            t, x, y, xvel, yvel = struct.unpack("<LHHhh",data[4:16])
            if addr not in state.peersfrom:
                state.sock.sendto('Said\0\0\0\0You are not connected, perhaps the server was restarted? Reconnect and try again.')
                continue
            player = state.peersfrom[addr]
            player.x = x
            player.y = y
            player.xvel = xvel * VELMUL
            player.yvel = yvel * VELMUL
            player.updated = state.long_to_time(t)
            state.broadcast("MOVU"+player.uid+data[4:16])
        elif data.startswith('Say '):
            player = state.peersfrom[addr]
            state.broadcast('Said'+player.uid+data[4:])
        elif data.startswith('Wear'):
            server_parse_wear(data, addr, now)
        else:
            print "Received invalid data",repr(data)

def server_do_exec():
    try:
        f = file('rapuexec.py','rb')
        data = f.read()
        f.close()
        exec data in globals(), locals()
    except Exception,e:
        print "Exception in exec:",e

def server_parse_drop(data, addr, now):
    x, y, what = struct.unpack('<HHB', data[4:9])
    player = state.peersfrom[addr]
    server_drop_shell(player, now)

def server_drop_shell(player, now):
    if player.shell==0:
        return
    shell = Shell(player.shell)
    shell.updated = now
    shell.uid = 'sh'+struct.pack("<H",random.randint(0,65535))

    player.update_position(now)

    shell.x = player.x
    shell.y = player.y
    shell.size = player.shell_size

    state.zone.append(shell)

    player.shell = 0
    player.shell_size = 32

    state.broadcast(player.info_str())
    state.broadcast(player.state_str())
    state.broadcast(shell.info_str())
    state.broadcast(shell.state_str())

def server_parse_drop_OLD(data, addr, now):
    x, y, what = struct.unpack('<HHB', data[4:9])
    shell = Shell(what)
    shell.updated = now
    shell.uid = 'sh'+struct.pack("<H",random.randint(0,65535))
    shell.x = x
    shell.y = y
    player = state.peersfrom[addr]

    shell.size = player.shell_size
    state.zone.append(shell)
    player.shell = 0
    player.shell_size = 32
    if verbose:
        check_hopping(player, now, x, y, 'before')
    player.update_position(now)
    if verbose:
        check_hopping(player, now, x, y, 'after')
    state.broadcast(player.info_str())
    state.broadcast(player.state_str())
    state.broadcast(shell.info_str())
    state.broadcast(shell.state_str())

def server_parse_wear(data, addr, now):
    player = state.peersfrom[addr]
    player.update_position(now)
    if player.hp<=0:
        state.sock.sendto('Said\0\0\0\0Ghosts cannot wear shells.', addr)
        return
    if len(data)<8:
        return
    what = state.remove_mob(data[4:8])
    if not what:
        return
    player.shell = what.spritenum
    player.shell_size = what.size
    state.broadcast('Worn'+player.uid+data[4:8]+chr(what.spritenum))

def bounded(orig, delt, lowbound, highbound):
    ret = orig+delt
    if ret<=lowbound:
        return orig
    if ret>=highbound:
        return orig
    return ret

class ClientState(State):
    mobs = {}
    epoch = 0
    my_uid = None
    sock = None
    me = None

    def __init__(cs):
        State.__init__(cs)

    def Connect(cs, servername, rapunimi, portnum=18116):
        cs.mobs = {}
        cs.buffer = []
        if not cs.my_uid:
            cs.my_uid = "%04x"%random.randint(0,65535)
        cs.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        cs.sock.connect((servername, portnum))
        cs.sockfd = cs.sock.fileno()
        cs.connected_time = time.time()
        pkt = 'CONN'+cs.my_uid
        if rapunimi:
            pkt += rapunimi
        cs.my_name = rapunimi
        cs.sock.send(pkt)
        cs.poll_all()

    def remove_mob(cs, uid):
        mob = State.remove_mob(cs, uid)
        if mob:
            del cs.mobs[uid]
            return mob
        return False

    def handle_net_input(cs, data):
        if data.startswith('MOVU'):
            mob_uid = data[4:8]
            if mob_uid not in cs.mobs:
                print "MOVU for unexpected unit",repr(mob_uid)
                cs.mobs[mob_uid] = mob = Mob()
                cs.zone.append(mob)
                mob.uid = mob_uid
            else:
                mob = cs.mobs[mob_uid]
            t, x, y, xvel, yvel = struct.unpack("<LHHhh", data[8:20])
            now = state.long_to_time(t)
            if verbose:
                check_hopping(mob, now, x, y)
            mob.updated = now
            mob.xvel = xvel * VELMUL
            mob.yvel = yvel * VELMUL
            mob.x = x
            mob.y = y
            # print "Setting", mob, x, y
        elif data.startswith('Said'):
            print "Got said:",repr(data)
            mob = cs.lookupcreat(data[4:8])
            out(unicode(mob.nimi)+": "+data[8:])
            mob.last_line = data[8:]
            renpy.exports.restart_interaction()
        elif data.startswith('DOTS'):
            if len(data)!=516:
                print "Improper dots length",repr(data),len(data)
            y = 0
            x = 0
            for c in data[4:]:
                cs.dots[y][x:x+8] = unpackbits(c)
                x += 8
                if x>=64:
                    x = 0
                    y += 1
        elif data.startswith('Info'):
            cs.parse_info(data)
        elif data.startswith('NOW='):
            current_tick, cs.bg_seed = struct.unpack('<LQ',data[4:16])
            now = time.time()
            # Assume 50% of transmission time was spent each direction.
            delt = now-cs.connected_time
            cs.epoch = (now-(delt*.5))-(current_tick/1024.)
        elif data.startswith('Nimi') and len(data)>8:
            peer = data[4:8]
            rapunimi = data[8:]
            cs.rename_mob(peer, rapunimi)
        elif data.startswith('Worn'):
            cs.parse_worn(data)
        elif data.startswith('Zone'):
            cs.parse_zone(data)
        else:
            print "Unhandled data",data

    def addvec(cs, xacc, yacc):
        player = cs.mobs[cs.my_uid]
        now = time.time()
        player.update_position(now)

        player.xvel = bounded(player.xvel, xacc, -2048, 2048)
        player.yvel = bounded(player.yvel, yacc, -2048, 2048)
        #print "player vel",player.xvel, player.yvel
        cs.sock.send("MOVE"+struct.pack("<LHHhh", cs.time_to_long(now),
            int(player.x), int(player.y), int(player.xvel*16), int(player.yvel*16)))

    def drop_shell(cs, quiet=False):
        me = cs.mobs[cs.my_uid]
        me.update_position(time.time())
        if me.shell:
            cs.sock.send("Drop"+struct.pack("<HHB", me.x, me.y, me.shell))
            me.shell = 0
            me.shell_size = 32
        elif not quiet:
            out("You aren't wearing a shell.")

    def wear_shell(cs, mob):
        me = cs.mobs[cs.my_uid]
        me.update_position(time.time())
        cs.sock.send("Wear"+mob.uid)

    def lookupcreat(cs, uid):
        if uid in cs.mobs:
            return cs.mobs[uid]
        cs.mobs[uid] = mob = Mob()
        cs.zone.append(mob)
        mob.uid = uid
        return mob
        
    def parse_info(cs, data):
        print "Debug: Client parsing info",repr(data)
        if len(data)<25:
            print "Malformed data",repr(data)
            return
        peer = data[4:8]
        mob = cs.lookupcreat(peer)
        mob.spr = None
        spr, r, g, b, size, stomach_capacity, stomach_full, hp, maxhp, shell_size, shell = struct.unpack("<BBBBHHHHHHB", data[8:25])
        mob.color = (r,g,b)
        mob.spritenum = spr
        mob.shell_size = shell_size
        mob.shell = shell
        mob.size = size
        mob.stomach_capacity = stomach_capacity
        mob.stomach_full = stomach_full
        mob.hp = hp
        mob.maxhp = maxhp
        rapunimi = data[25:]
        if rapunimi:
            cs.rename_mob(peer, rapunimi)
        if peer == cs.my_uid:
            cs.me = mob
            if renpy:
                renpy.exports.restart_interaction()

    def parse_worn(cs, data):
        if len(data)<13:
            return
        wearer = cs.mobs[data[4:8]]
        spritenum = ord(data[12])
        wearer.shell = spritenum
        wearer.shell_size = cs.mobs[data[8:12]].size
        print "Parse_worn setting shell size to",wearer.shell_size
        wearer.spr = None
        cs.remove_mob(data[8:12])
        if renpy:
            renpy.exports.restart_interaction()

    def parse_zone(cs, data):
        if len(data)<10:
            print "Malformed Zone command",repr(data)
            return
        ip = unpack("<BBBB",data[4:8])
        port = unpack("<H", data[8:10])
        cs.Connect("%d.%d.%d.%d"%ip, cs.rapunimi, portnum=port)

    def say(cs, s):
        cs.sock.send("Say "+s)

    def rename_mob(cs, peer, rapunimi):
        if peer not in cs.mobs:
            print "Name change from unregistered peer."
            return
        if cs.mobs[peer].nimi != rapunimi:
            cs.mobs[peer].spr = None
            out(repr(cs.mobs[peer].nimi)+" is now "+rapunimi)
            cs.mobs[peer].nimi = sanitized(rapunimi)

    def stop(cs):
        player = cs.mobs[cs.my_uid]
        now = time.time()
        player.update_position(now)
        player.xvel = player.yvel = 0
        cs.sock.send("MOVE"+struct.pack("<LHHhh", cs.time_to_long(now),
            int(player.x), int(player.y), int(player.xvel*16), int(player.yvel*16)))

    def poll(state):
        readable, discard, discard = select([state.sockfd],[],[],.001)
        if readable:
            data = cs.sock.recv(1024)
            try:
                state.handle_net_input(data)
            except Exception,e:
                print e,repr(data)
        now = time.time()
        if now>state.nexttick:
            update_simulation(state, now)

    def poll_all(state):
        while 1:
            readable, discard, discard = select([state.sockfd],[],[],.001)
            if readable:
                data = state.sock.recv(1024)
                try:
                    state.handle_net_input(data)
                except Exception,e:
                    print e,repr(data)
            else:
                now = time.time()
                if now>state.nexttick:
                    update_simulation(state, now)
                return
        
